# Generic makefile for a library
# Last edited on 2010-03-06 19:32:03 by stolfilocal

######################################################################
# To be included by the top-level Makefile in the library's directory.
# The caller must define
#  
#   ${LIBNAME}  name of library (WITHOUT the directory or the ".a")
#
# and may also define
#  
#   ${IGNORE} a list of ".h" or ".c" files that should not be compiled or installed.
#   ${OTHER_I_FLAGS}  additional -I flags for C compilation and dependency checks.
#   ${EXTRA_C_FLAGS} additional flags to compile C source files.

ifneq "/${LIBNAME}" "/"

# ----------------------------------------------------------------------
# Package components and directories

# Library file name:
LIBFILE = ${LIBNAME}.a

# Get list of sources:
HFILES := ${filter-out ${IGNORE}, ${wildcard *.h}}
CFILES := ${filter-out ${IGNORE}, ${wildcard *.c}}

# Derived object files:
HOFILES := ${subst .h,.ho,${HFILES}}
OFILES := ${subst .c,.o,${CFILES}}

# Directory with test programs and runs thereof:
TEST_DIR := tests

#---------------------------------------------------------------------
# Phony "make" targets:

.PHONY:: \
  all actions \
  build build-lib do-build-lib build-tests \
  check \
  tar-export tar-save \
  clean clean-lib clean-tests

#---------------------------------------------------------------------
# Default target: 

all: build-lib build-tests

#---------------------------------------------------------------------
# "make actions" performs the actions listed in ${ACTIONS}, if any

actions:
ifeq "/${ACTIONS}" "/"
	@echo '$${ACTIONS} is empty'
else
	${MAKE} ${ACTIONS}
endif

#---------------------------------------------------------------------
# Directories to search for #include files:

I_FLAGS := \
  -I. \
  ${OTHER_I_FLAGS}

#---------------------------------------------------------------------
# "make build" and "make build-lib" make sure that the ${DEPFILE}
# is up to date, and then recreate the object files from the source files
# as necessary.  "make build" and "make build-tests" also rebuild
# the test programs.

build: build-lib build-tests
  
build-lib: 
	${MAKE} do-build-lib
        
do-build-lib: ${LIBFILE}

CC := /usr/bin/gcc

C_FLAGS := \
  -Wall -Wundef \
  -Wpointer-arith \
  -Wmissing-prototypes \
  -ggdb \
  -fpcc-struct-return \
  -ffloat-store \
  -frounding-math

%.o: %.c
	${CC} -c ${C_FLAGS} ${EXTRA_C_FLAGS} ${I_FLAGS} $*.c
  
%.ho: %.h
	${CC} -o $*.ho -c ${C_FLAGS} ${EXTRA_C_FLAGS} ${I_FLAGS} -x c $*.h \
          || /bin/rm -f $*.ho
  
${LIBFILE}: ${HOFILES} ${OFILES}
	-rm -f $*.a
	ar crv $*.a ${OFILES}
	ranlib $*.a

build-tests: ${LIBFILE}
	if [[ -d ${TEST_DIR} ]]; then ( cd ${TEST_DIR} && ${MAKE} build ) fi

#---------------------------------------------------------------------
# "make check" runs the tests in the ${TEST_DIR} subdirectory.

check:
	if [[ -d ${TEST_DIR} ]]; then ( cd ${TEST_DIR} && ${MAKE} all ) fi

#---------------------------------------------------------------------
# "make clean" deletes all derived files.

clean: clean-lib clean-tests

clean-lib:
	-/bin/rm -f *.o *.ho *.a core
	if [[ -d ${TEST_DIR} ]]; then ( cd ${TEST_DIR} && ${MAKE} clean ) fi
  
clean-tests:
	cd ${TEST_DIR} && make clean
  
endif
# End of ${LIBNAME} section.
######################################################################
