/* See pst_imgsys.h 
**
** Created on 2005-10-01 by Jorge Stolfi, unicamp, <stolfi@dcc.unicamp.br>
** Based on the work of Rafael Saracchini, U.F.Fluminense.
** Last edited on 2008-11-10 20:26:57 by stolfi
** See the copyright and authorship notice at the end of this file. 
*/

#include <stdlib.h>
#include <assert.h>
#include <stdio.h>

#include <pst_imgsys.h>

#include <float_image.h>
#include <affirm.h>

double pst_imgsys_sol_change(double* Zold, double* Z, long int N);
  /* Returns the maximum of {abs(Zold[i]-Z[i])}, for {i = 0..N-1}. */

imgsys_t* pst_imgsys_new(long int N)
  {
    imgsys_t* S = (imgsys_t*)malloc(sizeof(imgsys_t));
    S->N = N;
    S->eq = (imgsys_equation_t*)malloc(sizeof(imgsys_equation_t)*N);
    return S;
  }

void pst_imgsys_solve
  ( imgsys_t* S, 
    double *Z, 
    long int max_iter, 
    double tol,
    int para, 
    int szero,
    bool_t debug,
    int indent
  )
  {
    long int N = S->N;
    long int i;

    /* Previous solution: */
    double *Zold = (double*)malloc(sizeof(double)*N);
    
    long int iter;
    for (iter = 0; iter < max_iter; iter++)
      {
        /* Another pass over all unknowns {Z[i]}: */
        for (i = 0; i < N; i++)
          {
            /* Store current solution in {Zold}: */
            Zold[i] = Z[i];
            /* Get equation {i}: */
            imgsys_equation_t *eqi = &(S->eq[i]);
            /* Compute {Z[i]} using equation {i}: */
            double sum = eqi->rhs; /* Right-hand side of equation. */
            double cf_i = 0.0; /* Coefficient of {Z[i]} in equation. */
            int k;
            for(k = 0; k < eqi->nt; k++)
              {
                /* Get hold of another unknown {Z[j] entering in equation {i}: */
                int j = eqi->ix[k];
                demand((j >= 0) && (j < N), "invalid variable index in system");
                double cf_j = eqi->cf[k];
                if (j == i) 
                  { /* This term uses {Z[i]}, store the coefficient: */
                    cf_i = cf_j;
                  }
                else if (cf_j != 0)
                  {
                    /* The unknown {Z[j]} is distinct from {Z[i]}. */
                    /* Get the appropriate value (new or old) of {Z[j]}: */
                    double Zj = (para && (j < i) ? Zold[j] : Z[j]);
                    /* Subtract from the right-hand side: */
                    sum = sum - Zj * cf_j;
                  }
              }
            
            /* Require that {eqi} depends on {Z[i]}: */
            demand(cf_i != 0.0, "system's matrix has a zero in the diagonal"); 
            
            /* Solve the equation: */
            Z[i] = sum / cf_i;
          }
          
        if (szero)
          { /* Normalize for zero sum: */
            double sum = 0;
            for (i = 0; i < N; i++) { sum += Z[i]; }
            double avg = sum/N;
            for (i = 0; i < N; i++) { Z[i] -= avg; }
          }
        iter++;
        double dif = pst_imgsys_sol_change(Zold, Z, N);
        if (debug)
          { fprintf(stderr, "%*s  iteration %3ld dif = %16.8f\n", indent, "", iter, dif); }
        if (dif <= tol) { return; }
      }
  }

double pst_imgsys_sol_change(double* Zold, double* Z, long int N)
  {
    int i;
    double max_dif = 0;
    long int max_i = 0;
    for(i = 0; i< N; i++)
      { double di = Z[i] - Zold[i];
        if(di < 0) { di = - di; }
        if(di > max_dif) { max_i = i; max_dif = di; }
      }
    return max_dif;
  }

    
void pst_imgsys_write(FILE *wr, imgsys_t *S, int NX, int NY)
  {
    long int i;
    assert(S->N == NX*NY);
    for(i = 0; i < S->N; i++)
      {
        /* Get equation {i}: */
        imgsys_equation_t *eqi = &(S->eq[i]);
        
        /* Compute indices of pixel corresponding to the unknown {Z[i]}: */
        int xi = i%NX; int yi = i/NX;
         
        /* Print indices of mai pixel: */
        fprintf(wr, "eq[%d][%d]:", xi, yi);

        /* Print the equation coefficients: */
        int k;
        for(k = 0; k < eqi->nt; k++)
          { /* Get hold of another unknown {Z[j]} that occurs in equation {i}: */
            long int j = eqi->ix[k];
            /* Compute indices of the corresponding pixel. */
            int xj = j%NX; int yj = j/NX;
            double cf_j = eqi->cf[k];
            if (k > 0) { fprintf(wr, " + "); }
            fprintf(wr, "%f*Z[%d][%d]", cf_j, xj, yj);
            demand((j >= 0) && (j < S->N), "invalid variable index in system");
          }
        /* Print the independent term: */
        fprintf(wr, " = %f", S->eq[i].rhs);
        fprintf(wr, "\n");
      }
    fflush(wr);
  }

void pst_imgsys_guess_sol_from_height_map(float_image_t *IZ, double *Z)
  {
    long int N = IZ->sz[1] * IZ->sz[2];
    long int i;
    for(i = 0; i < N; i++)
      { int x = i%IZ->sz[1];
        int y = i/IZ->sz[1];
        Z[i] = float_image_get_sample(IZ, 0, x, y);
      }
  }

void pst_imgsys_solution_to_height_map(double *Z, float_image_t *IZ)
  {
    long int N = IZ->sz[1] * IZ->sz[2];
    long int i;
    for(i = 0; i < N; i++)
      { int x = i%IZ->sz[1];
        int y = i/IZ->sz[1];
        float_image_set_sample(IZ, 0, x, y, Z[i]);
      }
  }

/*
**
** Copyright (C) Jorge Stolfi, Unicamp.
**
** Permission to use, copy, modify, and distribute this software and its
** documentation for any purpose and without fee is hereby granted, provided
** that the above copyright notice appear in all copies and that both that
** copyright notice and this permission notice appear in supporting
** documentation.  This software is provided "as is" without express or
** implied warranty. Neither the author nor its employers are liable to
** any damages which may result from its use.
*/
