/* imgsys - linear system solving for image-related problems.
**
** Created on 2005-12-04 by Jorge Stolfi, unicamp, <stolfi@dcc.unicamp.br>
** Based on the work of Rafael Saracchini, U.F.Fluminense.
** Last edited on 2010-02-03 13:27:12 by stolfi
** See the copyright and authorship notice at the end of this file. 
*/

#ifndef pst_imgsys_H
#define pst_imgsys_H

#include <float_image.h>

#define MAXCOEFS 5

typedef struct imgsys_equation_t
  { int nt;                  /* {nt} is the number of terms in the equation. */
    long int ix[MAXCOEFS];   /* {ix[k]} is the index of some unknown in the equation. */
    double cf[MAXCOEFS];  /* Coefficient of that unknown. */
    double rhs;              /* Right-hand side of equation. */
  } imgsys_equation_t;
  /* An {imgsys_equation_t} record represents a linear equation 
    with at most {MAXCOEFS} nonzero terms, namely
      {SUM {cf[k]*Z[ix[k]] : k = 0..nt-1} == rhs},
    where {Z[0..N-1]} are all the unknowns. */

typedef struct imgsys_t
  { long int N;
    imgsys_equation_t* eq;
  } imgsys_t;
  /* A {imgsys_t} represents {N} linear equations {eq[0..N-1]} on
    {N} unknowns {Z[0..N-1}. */

imgsys_t* pst_imgsys_new(long int N);
  /* Creates a new linear system with {N} equations on {N} unknowns.
    Doe snot initialize any coefficient or independent term. */

void pst_imgsys_solve
  ( imgsys_t* S, 
    double *Z, 
    long int max_iter, 
    double tol,
    int para, 
    int szero,
    bool_t debug,
    int indent
  );
  /* Solves system {S}, and stores the solution into {Z[0..N-1}}
     where {N = S->N}. 
     
     Uses an iterative method, and therefore assumes that the unknown
     {Z[i]} appears in some term of equation {S->eq[i]}, for {i =
     0..S->N-1}, and that its coefficient is ``large enough''. Upon
     entry, the {Z} vector must contain the starting guess. Executes
     at most {max_iter} iterations, but stops whenever two consecutive
     iterations do not change any variable by more than the tolerance
     {tol}.

     If {para} is TRUE, uses the ``parallel'' variant of the method
     (Jacobi), otherwise uses the sequential variant (Gauss-Seidel).
     If {szero == 1}, adjusts the solution so that it adds to zero,
     after each iteration.
     
     If {debug} is true, prints some debugging information.  All 
     debugging messages will be indented by {indent} spaces. */

void pst_imgsys_write(FILE *wr, imgsys_t *S, int NX, int NY);
  /* Writes the system {S} to stream {wr}.  The indices of unknowns and 
    equations are decomposed into pairs of pixel indices, assuming
    an image with {NX} columns and {NY} rows.  */
 
void pst_imgsys_guess_sol_from_height_map(float_image_t *IZ, double *Z);      
void pst_imgsys_solution_to_height_map(double *Z, float_image_t *IZ);
  /* Copies the values of the unknowns {Z[i]} from or to the
    elements of the height map {IZ[0,x,y]}. */

#endif

/*
**
** Copyright (C) Jorge Stolfi, Unicamp.
**
** Permission to use, copy, modify, and distribute this software and its
** documentation for any purpose and without fee is hereby granted, provided
** that the above copyright notice appear in all copies and that both that
** copyright notice and this permission notice appear in supporting
** documentation.  This software is provided "as is" without express or
** implied warranty. Neither the author nor its employers are liable to
** any damages which may result from its use.
*/
