/* See pst_slope_map.h */
/* Last edited on 2010-03-06 13:59:21 by stolfi */

#define _GNU_SOURCE
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>

#include <float_image.h>
#include <float_image_mscale.h>
#include <jsfile.h>
#include <r2.h>

#include <pst_basic.h>
#include <pst_imgsys.h>
#include <pst_slope_map.h>
#include <pst_weight_map.h>
#include <pst_height_map.h>

/* INTERNAL PROTOTYPES */

void pst_slope_map_write_debug_image(float_image_t *I, char *fileName, int indent);
  /* Writes the float image {I} in a human-readable format, to a file
    called "{fileName}" (which should include the extension ".fni").
    If {fileName} is "-", writes the image to standard output.
    Diagnostic messages are indented by {indent}. */

void pst_slope_map_write_debug_system(imgsys_t *S, char *fileName, int NX, int NY, int indent);
  /* Writes the system {S} to a file with the given {fileName} (which
    should include the extension ".sys").  The indices of unknowns and 
    equations are decomposed into pairs of pixel indices, assuming
    an image with {NX} columns and {NY} rows. Diagnostic
    messages are indented by {indent}.  */
    
#define get_sample float_image_get_sample
#define set_sample float_image_set_sample
 
/* IMPLEMENTATIONS */

r2_t pst_slope_map_get_pixel(float_image_t *IDZ, int x, int y)
  { demand(IDZ->sz[0] == 2, "wrong slope map depth");
    r2_t grd;
    float *p = float_image_get_sample_address(IDZ, 0, x, y);
    int axis;
    for (axis = 0; axis < 2; axis++)
      { grd.c[axis] = (*p); p += IDZ->st[0]; }
    return grd;
  }

void pst_slope_map_set_pixel(float_image_t *IDZ, int x, int y, r2_t *grd)
  { demand(IDZ->sz[0] == 2, "wrong slope map depth");
    float *p = float_image_get_sample_address(IDZ, 0, x, y);
    int axis;
    for (axis = 0; axis < 2; axis++)
      { (*p) = grd->c[axis]; p += IDZ->st[0]; }
  }

#define pst_slope_map_size_cutoff 1
  /* Do not recurse on images smaller than this. */

float_image_t *pst_slope_map_to_depth_map_recursive
  ( float_image_t *IDZ, 
    float_image_t *IW, 
    int level,
    bool_t hwavg,
    bool_t debugDZ,
    bool_t debugSys,
    bool_t debugZ,
    char *debugPrefix
  )
  {
    /* Get image size and check compatibility: */
    int NC = IDZ->sz[0]; assert(NC == 2);
    int NX = IDZ->sz[1];
    int NY = IDZ->sz[2];
    
    if (IW != NULL)
      { assert(IW->sz[0] == 1);
        assert(IW->sz[1] == NX);
        assert(IW->sz[2] == NY);
      }

    int indent = 2*level+2; /* Indentation for messages. */

    fprintf(stderr, "%*sEntering level %d (IDX,IDY = %d%d) ...\n", indent, "", level, NX, NY);
    char *fileName;

    /* Write slope maps, if asked for: */
    if (debugZ)
      { char *fileName;
        fileName = NULL; asprintf(&fileName, "%s-%02d-dZ.fni", debugPrefix, level);
        pst_slope_map_write_debug_image(IDZ, fileName, indent);
        free(fileName);
      }
    
    /* Allocate the height map: */
    float_image_t* IZ;
    
    /* Decide whether to use multi-scale: */
    bool_t small = ((NX < pst_slope_map_size_cutoff) && (NY < pst_slope_map_size_cutoff));
    bool_t atomic = ((NX == 1) || (NY == 1));
    bool_t direct = (small || atomic);
    
    /* Get the initial guess {IZ} for the heights: */
    if (direct)
      { /* Initialize {IZ} with zeros: */
        IZ = float_image_new(1, NX+1, NY+1);
        int x, y;
        for(y = 0; y < IZ->sz[2]; y++) 
          for(x = 0; x < IZ->sz[1]; x++) 
            {  set_sample(IZ, 0, x, y, 0.0); }
      }
    else
      { /* Initialize {IZ} with the solution to a coarser problem. */
      
        /* Shrink slope maps and weight map by half: */
        fprintf(stderr, "%*sShrinking slope maps ...\n", indent, "");
        float_image_t *JDZ = pst_slope_map_shrink(IDZ, IW);
        float_image_t *JW = (IW == NULL ? NULL : pst_weight_map_shrink(IW, hwavg));
        
        /* Compute the half-scale height map: */
        float_image_t *JZ = pst_slope_map_to_depth_map_recursive
          ( JDZ, JW, level+1, hwavg, debugDZ, debugSys, debugZ, debugPrefix );
        
        /* Expand the computed height map to double size: */
        fprintf(stderr, "%*sExpanding height map to %d%d ...\n", indent, "", NX+1, NY+1);
        IZ = pst_height_map_expand(JZ, NX+1, NY+1);
          
        /* Free the working storage: */
        float_image_free(JDZ);
        if (JW != NULL) { float_image_free(JW); }
        float_image_free(JZ);
      }

    /* Build the linear system: */
    long int N = (NX+1)*(NY+1);
    fprintf(stderr, "%*sBuilding linear system %ld%ld ...\n", indent, "", N, N);
    imgsys_t* S = pst_imgsys_new(N);
    pst_slope_map_build_integration_system(IDZ, IW, S);

    if (debugSys)
      { fileName = NULL; asprintf(&fileName, "%s-%02d.sys", debugPrefix, level);
        pst_slope_map_write_debug_system(S, fileName, IZ->sz[1], IZ->sz[2], indent);
        free(fileName);
      }

    /* Solve the system for the corner heights: */
    fprintf(stderr, "%*sSolving the system ...\n", indent, "");
    long int max_iter = 4000;
    double tol = 0.0000005;
    double *Z = (double*)malloc(sizeof(double)*S->N);
    int para = 0; /* 1 means parallel execution, 0 sequential. */
    int szero = 1; /* 1 means adjust sum to zero, 0 let it float. */
    pst_imgsys_guess_sol_from_height_map(IZ, Z);
    pst_imgsys_solve(S, Z, max_iter, tol, para, szero, debugSys, indent);
    pst_imgsys_solution_to_height_map(Z, IZ);
    
    /* Write height map, if asked for: */
    if (debugZ)
      { fileName = NULL; asprintf(&fileName, "%s-%02d-Z.fni", debugPrefix, level);
        pst_slope_map_write_debug_image(IZ, fileName, indent);
        free(fileName);
      }

    fprintf
      ( stderr, "%*sLeaving level %d (IZ = %lld%lld) ...\n", 
        indent, "", level, IZ->sz[1], IZ->sz[2]
      );

    return IZ;
  }

float_image_t *pst_slope_map_shrink_old(float_image_t *IDZ, float_image_t *IW)
  { 
    int NC = IDZ->sz[0];
    int NXI = IDZ->sz[1]; int NXJ = (NXI+1)/2;
    int NYI = IDZ->sz[2]; int NYJ = (NYI+1)/2;
    
    if (IW != NULL)
      { assert(IW->sz[0] == 1);
        assert(IW->sz[1] == NXI);
        assert(IW->sz[2] == NYI);
      }

    float_image_t *JDZ = float_image_new(NC, NXJ, NYJ);
    int xJ, yJ;
    for(yJ = 0; yJ < NYJ; yJ++)
      { for(xJ = 0; xJ < NXJ; xJ++)
          { int xI = 2*xJ, yI = 2*yJ;
            double w00 = 1.0, w01 = 1.0, w10 = 1.0, w11 = 1.0; /* Weights of each sample: */
            if (IW == NULL)
              { w00 = get_sample(IW, 0, xI, yI);
                w10 = ((xI+1 < NXI) ? get_sample(IW, 0, xI+1, yI) : w00);
                w01 = ((yI+1 < NYI) ? get_sample(IW, 0, xI, yI+1) : w00);
                w11 = ((xI+1 < NXI) && (yI+1 < NYI) ? get_sample(IW, 0, xI+1, yI+1) : w00);
              }
            double wtot = w00 + w01 + w10 + w11;
            if (wtot == 0.0) { w00 = w01 = w10 = w11 = 0.25; wtot = 1.0; }
            int c;
            for (c = 0; c < NC; c++)
              { /* Samples in original image (with border padding): */
                double d00 = get_sample(IDZ, c, xI, yI);
                double d10 = ((xI+1 < NXI) ? get_sample(IDZ, c, xI+1, yI) : d00);
                double d01 = ((yI+1 < NYI) ? get_sample(IDZ, c, xI, yI+1) : d00);
                double d11 = ((xI+1 < NXI) && (yI+1 < NYI) ? get_sample(IDZ, c, xI+1, yI+1) : d00);
                float v = (w00*d00 + w01*d01 + w10*d10 + w11*d11)/wtot;
                /* All three coordinate {X,Y,Z} get scaled, so the derivatives don't: */
                /* Set sample in shrunken image: */
                set_sample(JDZ, c, xJ, yJ, v);
              }
          }
      }
    return JDZ;
  }

float_image_t *pst_slope_map_shrink(float_image_t *IDZ, float_image_t *IW)
  { 
    return float_image_mscale_shrink(IDZ, IW, 2);
  }

void pst_slope_map_build_integration_system(float_image_t *IDZ, float_image_t *IW, imgsys_t* S)
  {

    /* Get/check the sizes of the slope maps: */
    int NC = IDZ->sz[0]; assert(NC == 2);
    int NX = IDZ->sz[1];
    int NY = IDZ->sz[2];
    
    if (IW != NULL)
      { assert(IW->sz[0] == 1);
        assert(IW->sz[1] == NX);
        assert(IW->sz[2] == NY);
      }
    
    /* Check the size of the system: */
    int NH = (NX+1)*(NY+1); /* Number of unknowns (heights). */
    assert(S->N == NH);
    
    /* Conceptually, we define two quadratic mismatch terms
      {qx[x,y](h)} and {qy[x,y](h)} for each edge in the pixel grid.
      
      The term {qx[x,y](h)} is the square of the difference between
      the mean slope along the edge from {(x,y)} to {(x+1,y)}, as
      estimated from the slope map {IDZ}, and the mean slope computed
      by numerical difference of the (unknown) heights at the two
      endpoints of that edge. This terms exists for all {x} in {0..NX-1}
      and all {y} in {0..NY}.
      
        { qx[x,y](h) = (dx[x,y] - (- h[k00] + h[k10]))^2 }

      where {h[k00]} and {h[k10]} are the {h} variables
      corresponding to heights {z[x,y]} and {z[x+1,y]}, respectively;
      and {dx} is the interpolated slope
      
        { dx[x,y] = (wpm*dxpm + wpp*dxpp)/wx[x,y] }

      where {wpm == IW[0,x,y-1]}, {wpp == IW[0,x,y]}, {wx[x,y] == wpm+wpp},
      {dxpm == IDZ[0,x,y-1]}, and {dxpp == IDZ[0,x,y]}.
      
      The term {qy[x,y](h)} is similarly defined for edges parallel to
      the Y axis. It exists for all {x} in {0..NX} and all {y} in
      {0..NY-1}.
      
        { qy[x,y](h) = (dy[x,y] - (- h[k00] + h[k01]))^2 }
      
        { dy[x,y] = (wmp*dxmp + wpp*dxpp)/wy[x,y] }

      where {wmp == IW[0,x,y-1]}, {wy[x,y] == wmp+wpp},
      {dymp == IDZ[1,x-1,y]}, and {dypp == IDZ[1,x,y]}.

      We want to minimize the sum {Q(h)} of all terms {qx[x,y](h)} and 
      {qy[x,y](h)}, each weighted by the corresponding edge weight
      {wx[x,y]} or {wy[x,y]}. 
      
      Note that if the slope map {IDZ} is zero, we can anihilate all
      those terms (and therefore Q) by setting all heights equal to 1.
      Therefore, if the system is to be solved iteratively, one must
      take care to exclude this solution.
       
      Minimizing {Q} is equivalent to finding the {h} vector that
      anihilates the gradient {G} of {Q}. The gradient of {Q} is an
      affine function of {h}, namely {G(h) = A h - b}. Each term
      {wx[x,y]*qx[x,y](h)} contributes to two components of the
      gradient, namely those with indices {k00,k10}. Similarly, each
      term {wy[x,y]*qy[x,y]} contributes to two components, with
      indices {k00,k01}. Thus, to build the system, we add those
      contributions to the appropriate elements of {A} and {b}.
     
      OOPS. If the weights are all equal, this system is insensitive 
      to any first-degree component of the height field, except 
      along the edges. But this is OK, I guess...
    */
    
    auto long int ind_h(int x, int y); 
      /* This function computes the index into the solution vector {h}
        corresponding to vertex {[x,y]} of the images. */
    
    long int ind_h(int x, int y) { return x + y*(NX+1); }
    
    /* Make sure that we can build the system: */
    assert(MAXCOEFS >= 5);

    /* Enumerate the vertices of {IDX,IDY}, i.e. elems of the height map: */
    int x, y;
    for(y = 0; y <= NY; y++)
      { for(x = 0; x <= NX; x++)
          { /* Get the slopes and relative weights of the pixels incident to {(x,y)}: */
            double dxmm = 0.0, dxmp = 0.0, dxpm = 0.0, dxpp = 0.0;
            double dymm = 0.0, dymp = 0.0, dypm = 0.0, dypp = 0.0;
            double wmm = 0.0, wmp = 0.0, wpm = 0.0, wpp = 0.0;

            if ((x > 0) && (y > 0))
              { dxmm = get_sample(IDZ, 0, x-1, y-1);
                dymm = get_sample(IDZ, 1, x-1, y-1);
                wmm = (IW == NULL ? 1.0 : get_sample(IW, 0, x-1, y-1));
                assert(wmm >= 0.0);
              }

            if ((x > 0) && (y < NY))
              { dxmp = get_sample(IDZ, 0, x-1, y);
                dymp = get_sample(IDZ, 1, x-1, y);
                wmp = (IW == NULL ? 1.0 : get_sample(IW, 0, x-1, y));
                assert(wmp >= 0.0);
              }

            if ((x < NX) && (y > 0))
              { dxpm = get_sample(IDZ, 0, x, y-1);
                dypm = get_sample(IDZ, 1, x, y-1);
                wpm = (IW == NULL ? 1.0 : get_sample(IW, 0, x, y-1));
                assert(wpm >= 0.0);
              }

            if ((x < NX) && (y < NY))
              { dxpp = get_sample(IDZ, 0, x, y);
                dypp = get_sample(IDZ, 1, x, y);
                wpp = (IW == NULL ? 1.0 : get_sample(IW, 0, x, y));
                assert(wpp >= 0.0);
              }
              
            assert(! isnan(wmm));
            assert(! isnan(wmp));
            assert(! isnan(wpm));
            assert(! isnan(wpp));
           
            /* Get the weights of the edges incident to {(x,y)}: */
            double wxm = wmm + wmp;
            double wxp = wpm + wpp;
            double wym = wmm + wpm;
            double wyp = wmp + wpp; 
            
            /* Compute the edge slopes times the edge weights: */
            double wdxm = 0.0, wdxp = 0.0, wdym = 0.0, wdyp = 0.0;

            double wtot = wxm + wxp + wym + wyp;
            if (wtot == 0.0)
              { /* Height {z[x,y]} is loose. Make it be the average of its neighbors: */
                wxm = (x <= 0  ? 0.0 : 1.0);
                wxp = (x >= NX ? 0.0 : 1.0);
                wym = (y <= 0  ? 0.0 : 1.0);
                wyp = (y >= NY ? 0.0 : 1.0);
                wtot = wxm + wxp + wym + wyp;
                /* Leave {wdxm,wdxp,wdym,wdyp} all zero. */
              }
            else
              { /* Compute {wdxm,wdxp,wdym,wdyp} from given slopes: */
                wdxm = wmm*dxmm + wmp*dxmp;
                wdxp = wpm*dxpm + wpp*dxpp;
                wdym = wmm*dymm + wpm*dypm;
                wdyp = wmp*dymp + wpp*dypp;
              }
            assert(wtot > 0.0);
            assert(! isnan(wdxm));
            assert(! isnan(wdxp));
            assert(! isnan(wdym));
            assert(! isnan(wdyp));
                
            /* Get the index {k} into {h} of {IZ[x,y]} and its cardinal neighbors: */
            long int k = ind_h(x, y); 
            long int kxm = ind_h(x-1, y); 
            long int kxp = ind_h(x+1, y); 
            long int kym = ind_h(x, y-1); 
            long int kyp = ind_h(x, y+1); 
            
            /* Assemble the equation for height {z[x,y]}: */
            imgsys_equation_t *eqk = &(S->eq[k]);
            int nt = 0; /* Number of terms in equation. */
            eqk->rhs = 0.0;
            eqk->ix[nt] = k;  eqk->cf[nt] = 1.00; nt++;
            if (wxm != 0.0) 
              { eqk->ix[nt] = kxm; eqk->cf[nt] = -wxm/wtot; eqk->rhs += +wdxm/wtot; nt++; }
            if (wxp != 0.0) 
              { eqk->ix[nt] = kxp; eqk->cf[nt] = -wxp/wtot; eqk->rhs += -wdxp/wtot; nt++; }
            if (wym != 0.0) 
              { eqk->ix[nt] = kym; eqk->cf[nt] = -wym/wtot; eqk->rhs += +wdym/wtot; nt++; }
            if (wyp != 0.0) 
              { eqk->ix[nt] = kyp; eqk->cf[nt] = -wyp/wtot; eqk->rhs += -wdyp/wtot; nt++; }
            eqk->nt = nt;
          }
      }
  }
  
void pst_slope_map_write_debug_image(float_image_t *I, char *fileName, int indent)
  { FILE* wr = open_write(fileName, FALSE);
    float_image_write(wr, I);
    if (wr == stdout) { fflush(wr); } else { fclose(wr); }
    fprintf(stderr, "\n");
  }

void pst_slope_map_write_debug_system(imgsys_t *S, char *fileName, int NX, int NY, int indent)
  { fprintf(stderr, "%*sWriting %s ...", indent, "", fileName);
    FILE* wr = fopen(fileName, "wt");
    assert(wr != NULL); 
    pst_imgsys_write(wr, S, NX, NY);
    if (wr == stdout) { fflush(wr); } else { fclose(wr); }
    fprintf(stderr, "\n");
  }
