#ifndef pst_slope_map_H
#define pst_slope_map_H

/* pst_slope_map.h -- procedures for working with slope maps. */
/* Last edited on 2009-02-23 23:14:38 by stolfi */

#include <bool.h>
#include <r2.h>

#include <float_image.h>
#include <pst_imgsys.h>

/* SLOPE MAPS
  
  A /slope map/ is a two-channel float-valued image of some
  height field {Z(X,Y)}, where the value of each pixel is the 
  gradient of that height --- that is, its derivatives
  {dZ/dX} and {dZ/dY}.
  
  !!! TO DO !!!
  
  !!! Make {pst_slope_map_build_integration_system_1} accept a weight map, too.
  
  !!! Generalize to use a 2-channel weight map, with separate
  reliability weights for the X and Y component of the gradient. 
  
*/

r2_t pst_slope_map_get_pixel(float_image_t *IDZ, int x, int y);
  /* Extracts the derivatives {dZ/dX} and {dZ/dY} from channels 0 and
    1 of the pixel in column {x}, row {y} of {IDZ}, and returns them
    as a {r2_t} gradient vector. */

void pst_slope_map_set_pixel(float_image_t *IDZ, int x, int y, r2_t *grd);
  /* Stores the derivatives {dZ/dX} and {dZ/dY}, taken from the
    gradient vector {grd} into channels 0 and 1 of the pixel in column
    {x}, row {y} of {IDZ}. */

float_image_t *pst_slope_map_shrink(float_image_t *IDZ, float_image_t *IW);
  /* Given a slope map {IDZ}, containing the derivative of a height
    function {Z} along X and Y axes, returns another slope map {JDZ},
    with half the size as {IDZ}, containing the derivatives of a
    version {JZ} of {IZ} with both dimensions and heights scaled by
    half.
    
    If {IW} is not null, it should be a monochromatic image with the
    same size as {IDZ}. Each element {IW[0,x,y]} (which must be
    non-negative) is interpreted as the relative reliability of the
    slopes {IDZ[c,x,y]}, and is used as a weight in the averaging
    of the slopes.
    
    Pixels of {IDZ} that have zero reliability in {IW} are ignored in
    the averaging (unless all pixels to be averaged have zero
    reliability, in which case all those reliabilities are taken to be
    1). */

float_image_t *pst_slope_map_shrink_old(float_image_t *IDZ, float_image_t *IW);
  /* Older version of {pst_slope_map_shrink}.
    !!! Delete after the new one is tested. !!! */

/* COMPUTING HEIGHTS FROM SLOPES */
      
float_image_t *pst_slope_map_to_depth_map_recursive
  ( float_image_t *IDZ, 
    float_image_t *IW, 
    int level, 
    bool_t hwavg,
    bool_t debugDZ,
    bool_t debugSys,
    bool_t debugZ,
    char *debugPrefix
  );
  /* Computes a depth map {IZ} given the slope map {IDZ}, where channel
    0 is {dZ/dX} and channel 1 is {dZ/dY}. The {IZ} image will be
    bigger than {IDZ} by one row and one column.
    
    If {IW} is not NULL, it is taken to be a single-channel weight map, 
    with the same dimensions as the slope map.
    
    The image is computed by solving a set of linear equations.
    
    The linear system is solved by the Gauss-Jordan iterative method.
    The initial guess for this method is obtained by scaling down the
    slope maps {IDZ} by 1/2 in all three axes, computing the
    corresponding height field recursively, and un-scaling the result.
    The {level} parameter indicates the deph of the recursion.
    The recursion stops when the maps are small enough.
    
    The slope maps are reduced by arithmetic averaging the original
    slope values that correspond to the same reduced pixel (see
    {pst_slope_map_shrink}). The weight maps are reduced by arithmetic
    averagin if {hwavg} is false, by harmonic averaging if {hwavg} is
    true (see {pst_weight_map_shrink}).
    
    If {debugZ=TRUE}, the procedure writes images containing the
    height maps obtained at each scale, to files called
    "{XXX}-{level}-Z.fni" where {XXX==debugPrefix}. If {debugDZ=TRUE},
    it dumps also the slope maps at each scale, to files
    "{XXX}-{level}-dZ.fni". If {debugSys=TRUE}, it also writes the
    linear system of each scale to files "{XXX}-{level}.sys".*/

/* THE INTEGRATION SYSTEM 
 */

void pst_slope_map_build_integration_system(float_image_t *IDZ, float_image_t *IW, imgsys_t* S);
  /*
    Stores into the {imgsys_t} system {S} the linear equations that
    yield the height map {IZ} given a slope map {IDZ}.

    The value of{IDZ[0,x,y]} must be the X derivative {dZ/dX} of the heigh map,
    averaged over the pixel {[x,y]} The value of {IDZ[1,x,y]} must
    be the Y derivative {dZ/dY} averaged over the same pixel.

    The image {IW}, if not {NULL}, must have the same number of rows
    and columns as the slope map {IDZ}. The value of {IW[0,x,y]} is
    interpreted as the relative weight or reliability of the slope
    data {IDZ[c,x,y]}. The absolute value of the weights is not
    relevant, only their ratios. The smaller {IW[0,x,y]} is, compared
    to the weights of neighboring pixels, the smaller will be the
    influence of the slopes {IDZ[c,x,y]} on the computed heights.
    
    The weights must be non-negative. To prevent indeterminacies, zero
    weights are implicitly replaced by some very small weight.

    The system has the form {A h = b} where {A} is a known matrix, {h}
    is the unknown solution vector, and {b} is a known vector. The
    unknowns are the heights (Z coordinates) at the CORNERS of all
    pixels in {IDZ}. Therefore, if the slope map {IDZ} has {NX} columns
    and {NY} rows, the system has {NH = (NX+1)*(NY+1)} equations and
    unknowns.

    More precisely, {h[x + (NX+1)*y} is the height at the integer point
    {(x,y)}, not at the pixel center {(x+0.5,y+0.5)}, for all {x} in
    {0..NX} and all {y} in {0..NY}.
    
    If all weights are equal, the equations essentially state that the
    Laplacian of the height, as computed from the (unknown) height map
    {IZ}, is equal to the Laplacian computed from the (known) slope
    maps {IDZ}. Both Laplacians are computed by finite-difference
    formulas. Other equations, based on equality of slopes, are used
    along the edges and where the weights are zero.
    
    If all weights are non-zero, the system has a one-dimensional
    family of solutions, since the homogeneous system {A h = 0} 
    is satisfied by {h[k]==1} for all {k}. */

#endif
